"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.declareModuleBindings = exports.declareModule = exports.bindModules = exports.bindModule = void 0;
const ditox_1 = require("./ditox");
const utils_1 = require("./utils");
/**
 * Binds the dependency module to the container
 * @param container - Dependency container.
 * @param moduleDeclaration - Declaration of the dependency module.
 * @param options - Options for module binding.
 *
 * @example
 * ```ts
 * bindModule(container, LOGGER_MODULE);
 * ```
 */
function bindModule(container, moduleDeclaration, options) {
    const { token, imports, factory, beforeBinding, afterBinding } = moduleDeclaration;
    const exports = moduleDeclaration.exports;
    const scope = options === null || options === void 0 ? void 0 : options.scope;
    if (beforeBinding) {
        beforeBinding(container);
    }
    if (imports) {
        bindModules(container, imports);
    }
    const exportedValueTokens = new Set();
    if (exports) {
        const keys = Object.keys(exports);
        keys.forEach((valueKey) => {
            const valueToken = exports[valueKey];
            if (valueToken) {
                exportedValueTokens.add(valueToken);
                container.bindFactory(valueToken, (0, utils_1.injectable)((module) => module[valueKey], token), { scope });
            }
        });
    }
    container.bindFactory(token, factory, {
        scope,
        onRemoved: (module) => {
            if (module.destroy) {
                module.destroy();
            }
            exportedValueTokens.forEach((valueToken) => container.remove(valueToken));
            exportedValueTokens.clear();
        },
    });
    if (afterBinding) {
        afterBinding(container);
    }
}
exports.bindModule = bindModule;
/**
 * Binds dependency modules to the container
 *
 * @param container - Dependency container for binding
 * @param modules - Array of module binding entries: module declaration or `{module: ModuleDeclaration, options: BindModuleOptions}` objects.
 */
function bindModules(container, modules) {
    modules.forEach((entry) => {
        if ('module' in entry) {
            bindModule(container, entry.module, entry.options);
        }
        else {
            bindModule(container, entry);
        }
    });
}
exports.bindModules = bindModules;
/**
 * Declares a module binding
 *
 * @param declaration - a module declaration
 * @param declaration.token - optional field
 *
 *  @example
 * ```ts
 * const LOGGER_MODULE = declareModule<LoggerModule>({
 *   factory: (container) => {
 *     const transport = container.resolve(TRANSPORT_TOKEN).open();
 *     return {
 *       logger: { log: (message) => transport.write(message) },
 *       destroy: () => transport.close(),
 *     }
 *   },
 *   exports: {
 *     logger: LOGGER_TOKEN,
 *   },
 * });
 * ```
 */
function declareModule(declaration) {
    var _a;
    return { ...declaration, token: (_a = declaration.token) !== null && _a !== void 0 ? _a : (0, ditox_1.token)() };
}
exports.declareModule = declareModule;
/**
 * Declares bindings of several modules
 *
 * @param modules - module declaration entries
 */
function declareModuleBindings(modules) {
    return declareModule({
        factory: () => ({}),
        imports: modules,
    });
}
exports.declareModuleBindings = declareModuleBindings;
//# sourceMappingURL=modules.js.map