/**
 * Creates a new binding token.
 * @param description - Token description for better error messages.
 */
export function token(description) {
    return { symbol: Symbol(description) };
}
export function optional(token, optionalValue) {
    return {
        symbol: token.symbol,
        isOptional: true,
        optionalValue,
    };
}
/**
 * ResolverError is thrown by the resolver when a token is not found in a container.
 */
export class ResolverError extends Error {
    constructor(message) {
        super(message);
        Object.setPrototypeOf(this, new.target.prototype);
        this.name = 'ResolverError';
    }
}
/** @internal */
export const CONTAINER = token('ditox.Container');
/** @internal */
export const PARENT_CONTAINER = token('ditox.ParentContainer');
/** @internal */
export const RESOLVER = token('ditox.Resolver');
/** @internal */
const NOT_FOUND = Symbol();
/** @internal */
export const FAKE_FACTORY = () => {
    throw new Error('FAKE_FACTORY');
};
/** @internal */
const DEFAULT_SCOPE = 'singleton';
/** @internal */
export const FACTORIES_MAP = token('ditox.FactoriesMap');
/** @internal */
function getScope(options) {
    var _a;
    return (_a = options === null || options === void 0 ? void 0 : options.scope) !== null && _a !== void 0 ? _a : DEFAULT_SCOPE;
}
/** @internal */
function getOnRemoved(options) {
    return options.scope === undefined ||
        options.scope === 'scoped' ||
        options.scope === 'singleton'
        ? options.onRemoved
        : undefined;
}
/** @internal */
function isInternalToken(token) {
    return (token.symbol === CONTAINER.symbol ||
        token.symbol === PARENT_CONTAINER.symbol ||
        token.symbol === RESOLVER.symbol);
}
/**
 * Creates a new dependency container.
 *
 * Container can have an optional parent to chain token resolution. The parent is used in case the current container does not have a registered token.
 *
 * @param parentContainer - Optional parent container.
 */
export function createContainer(parentContainer) {
    const values = new Map();
    const factories = new Map();
    const container = {
        bindValue(token, value) {
            if (isInternalToken(token)) {
                return;
            }
            values.set(token.symbol, value);
        },
        bindFactory(token, factory, options) {
            if (isInternalToken(token)) {
                return;
            }
            factories.set(token.symbol, { factory, options });
        },
        remove(token) {
            var _a;
            if (isInternalToken(token)) {
                return;
            }
            const options = (_a = factories.get(token.symbol)) === null || _a === void 0 ? void 0 : _a.options;
            if (options) {
                executeOnRemoved(token.symbol, options);
            }
            values.delete(token.symbol);
            factories.delete(token.symbol);
        },
        removeAll() {
            factories.forEach((context, tokenSymbol) => {
                if (context.options) {
                    executeOnRemoved(tokenSymbol, context.options);
                }
            });
            values.clear();
            factories.clear();
            bindInternalTokens();
        },
        hasToken(token) {
            var _a;
            return (values.has(token.symbol) ||
                factories.has(token.symbol) ||
                ((_a = parentContainer === null || parentContainer === void 0 ? void 0 : parentContainer.hasToken(token)) !== null && _a !== void 0 ? _a : false));
        },
        get(token) {
            const value = resolver(token, container);
            if (value !== NOT_FOUND) {
                return value;
            }
            if (token.isOptional) {
                return token.optionalValue;
            }
            return undefined;
        },
        resolve(token) {
            var _a;
            const value = resolver(token, container);
            if (value !== NOT_FOUND) {
                return value;
            }
            if (token.isOptional) {
                return token.optionalValue;
            }
            throw new ResolverError(`Token "${(_a = token.symbol.description) !== null && _a !== void 0 ? _a : ''}" is not provided`);
        },
    };
    function resolver(token, origin) {
        const value = values.get(token.symbol);
        const hasValue = value !== undefined || values.has(token.symbol);
        if (hasValue && origin === container) {
            return value;
        }
        const factoryContext = factories.get(token.symbol);
        if (factoryContext && factoryContext.factory !== FAKE_FACTORY) {
            const scope = getScope(factoryContext.options);
            switch (scope) {
                case 'singleton': {
                    if (hasValue) {
                        return value;
                    }
                    else if (parentContainer === null || parentContainer === void 0 ? void 0 : parentContainer.hasToken(token)) {
                        break;
                    }
                    else {
                        // Cache the value in the same container where the factory is registered.
                        const value = factoryContext.factory(container);
                        container.bindValue(token, value);
                        return value;
                    }
                }
                case 'scoped': {
                    // Create a value within the origin container and cache it.
                    const value = factoryContext.factory(origin);
                    origin.bindValue(token, value);
                    if (origin !== container) {
                        // Bind a fake factory with actual options to make onRemoved() works.
                        origin.bindFactory(token, FAKE_FACTORY, factoryContext.options);
                    }
                    return value;
                }
                case 'transient': {
                    // Create a value within the origin container and don't cache it.
                    return factoryContext.factory(origin);
                }
            }
        }
        if (hasValue) {
            return value;
        }
        const parentResolver = parentContainer === null || parentContainer === void 0 ? void 0 : parentContainer.get(RESOLVER);
        if (parentResolver) {
            return parentResolver(token, origin);
        }
        return NOT_FOUND;
    }
    function executeOnRemoved(tokenSymbol, options) {
        const onRemoved = getOnRemoved(options);
        if (onRemoved) {
            const value = values.get(tokenSymbol);
            if (value !== undefined || values.has(tokenSymbol)) {
                onRemoved(value);
            }
        }
    }
    function bindInternalTokens() {
        values.set(CONTAINER.symbol, container);
        values.set(RESOLVER.symbol, resolver);
        values.set(FACTORIES_MAP.symbol, factories);
        if (parentContainer) {
            values.set(PARENT_CONTAINER.symbol, parentContainer);
        }
    }
    bindInternalTokens();
    return container;
}
//# sourceMappingURL=ditox.js.map