import { Container, Token } from './ditox';
declare type AnyObject = Record<string, any>;
declare type EmptyObject = Record<string, never>;
declare type ModuleController = {
    /** Dispose the module and clean its resources */
    destroy?: () => void;
};
/**
 * Dependency module
 *
 * @example
 * ```ts
 * type LoggerModule = Module<{
 *   logger: Logger;
 * }>;
 * ```
 */
export declare type Module<ModuleProps extends AnyObject = EmptyObject> = ModuleController & ModuleProps;
declare type GetModuleProps<T> = T extends Module<infer Props> ? Props : never;
/**
 * Description how to bind the module in declarative way.
 *
 * @example
 * ```ts
 * const LOGGER_MODULE: ModuleDeclaration<LoggerModule> = {
 *   token: LOGGER_MODULE_TOKEN,
 *   factory: (container) => {
 *     const transport = container.resolve(TRANSPORT_TOKEN).open();
 *     return {
 *       logger: { log: (message) => transport.write(message) },
 *       destroy: () => transport.close(),
 *     }
 *   },
 *   exports: {
 *     logger: LOGGER_TOKEN,
 *   },
 * };
 * ```
 */
export declare type ModuleDeclaration<T extends Module<AnyObject>> = {
    /** Token for the module */
    token: Token<T>;
    /** Modules for binding  */
    imports?: ReadonlyArray<ModuleBindingEntry>;
    /** Factory of the module */
    factory: (container: Container) => T;
    /** Dictionary of module properties which are bound to tokens. */
    exports?: {
        [K in keyof GetModuleProps<T>]?: Token<GetModuleProps<T>[K]>;
    };
    /** Callback could be used to prepare an environment. It is called before binding the module. */
    beforeBinding?: (container: Container) => void;
    /** Callback could be used to export complex dependencies from the module. It is called after binding the module.  */
    afterBinding?: (container: Container) => void;
};
/**
 * Options for module binding.
 *
 * `scope` types:
 *   - `singleton` - **This is the default**. The module is created and cached by the container which registered the factory.
 *   - `scoped` - The module is created and cached by the container which starts resolving.
 */
export declare type BindModuleOptions = {
    scope?: 'scoped' | 'singleton';
};
export declare type ModuleBindingEntry = ModuleDeclaration<AnyObject> | {
    module: ModuleDeclaration<AnyObject>;
    options: BindModuleOptions;
};
/**
 * Binds the dependency module to the container
 * @param container - Dependency container.
 * @param moduleDeclaration - Declaration of the dependency module.
 * @param options - Options for module binding.
 *
 * @example
 * ```ts
 * bindModule(container, LOGGER_MODULE);
 * ```
 */
export declare function bindModule<T extends Module<AnyObject>>(container: Container, moduleDeclaration: ModuleDeclaration<T>, options?: BindModuleOptions): void;
/**
 * Binds dependency modules to the container
 *
 * @param container - Dependency container for binding
 * @param modules - Array of module binding entries: module declaration or `{module: ModuleDeclaration, options: BindModuleOptions}` objects.
 */
export declare function bindModules(container: Container, modules: ReadonlyArray<ModuleBindingEntry>): void;
/**
 * Declares a module binding
 *
 * @param declaration - a module declaration
 * @param declaration.token - optional field
 *
 *  @example
 * ```ts
 * const LOGGER_MODULE = declareModule<LoggerModule>({
 *   factory: (container) => {
 *     const transport = container.resolve(TRANSPORT_TOKEN).open();
 *     return {
 *       logger: { log: (message) => transport.write(message) },
 *       destroy: () => transport.close(),
 *     }
 *   },
 *   exports: {
 *     logger: LOGGER_TOKEN,
 *   },
 * });
 * ```
 */
export declare function declareModule<T>(declaration: Omit<ModuleDeclaration<T>, 'token'> & Partial<Pick<ModuleDeclaration<T>, 'token'>>): ModuleDeclaration<T>;
/**
 * Declares bindings of several modules
 *
 * @param modules - module declaration entries
 */
export declare function declareModuleBindings(modules: ReadonlyArray<ModuleBindingEntry>): ModuleDeclaration<Module>;
export {};
