/**
 * Checks if a value is the token
 */
export function isToken(value) {
    return (value !== undefined &&
        value !== null &&
        typeof value === 'object' &&
        'symbol' in value &&
        typeof value.symbol === 'symbol');
}
/**
 * Rebinds the array by the token with added new value.
 * @param container - Dependency container.
 * @param token - Token for an array of values.
 * @param value - New value which is added to the end of the array.
 */
export function bindMultiValue(container, token, value) {
    var _a;
    const prevValues = (_a = container.get(token)) !== null && _a !== void 0 ? _a : [];
    const nextValues = [...prevValues, value];
    container.bindValue(token, nextValues);
}
/**
 * Tries to resolve a value by the provided token.
 *
 * If an argument is an object which has tokens as its properties,
 * then returns an object containing resolved values as properties.

 * If a token is not found, then `undefined` value is used.
 *
 * @example
 * ```ts
 * const value = tryResolveValue(container, tokenA);
 * console.log(value); // 1
 *
 * const props = tryResolveValue(container, {a: tokenA, b: tokenB});
 * console.log(props); // {a: 1, b: 2}
 * ```
 */
export function tryResolveValue(container, token) {
    if (isToken(token)) {
        return container.get(token);
    }
    const obj = {};
    Object.keys(token).forEach((key) => (obj[key] = container.get(token[key])));
    return obj;
}
/**
 * Returns an array of resolved values or objects with resolved values.
 *
 * If an item of the array is an object which has tokens as its properties,
 * then returns an object containing resolved values as properties.

 * If a token is not found, then `undefined` value is used.
 *
 * @example
 * ```ts
 * const items1 = tryResolveValues(container, tokenA);
 * console.log(items1); // [1]
 *
 * const items2 = tryResolveValues(container, tokenA, {a: tokenA, b: tokenB});
 * console.log(items2); // [1, {a: 1, b: 2}]
 * ```
 */
export function tryResolveValues(container, ...tokens) {
    return tokens.map((item) => tryResolveValue(container, item));
}
/**
 * Resolves a value by the provided token.
 *
 * If an argument is an object which has tokens as its properties,
 * then returns an object containing resolved values as properties.

 * If a value is not found by the token, then `ResolverError` is thrown.
 *
 * @example
 * ```ts
 * const value = resolveValue(container, tokenA);
 * console.log(value); // 1
 *
 * const props = resolveValue(container, {a: tokenA, b: tokenB});
 * console.log(props); // {a: 1, b: 2}
 * ```
 */
export function resolveValue(container, token) {
    if (isToken(token)) {
        return container.resolve(token);
    }
    const obj = {};
    Object.keys(token).forEach((key) => (obj[key] = container.resolve(token[key])));
    return obj;
}
/**
 * Returns an array of resolved values or objects with resolved values.
 *
 * If an item of the array is an object which has tokens as its properties,
 * then returns an object containing resolved values as properties.

 * If a token is not found, then `ResolverError` is thrown.
 *
 * @example
 * ```ts
 * const items1 = resolveValues(container, tokenA);
 * console.log(items1); // [1]
 *
 * const items2 = resolveValues(container, tokenA, {a: tokenA, b: tokenB});
 * console.log(items2); // [1, {a: 1, b: 2}]
 * ```
 */
export function resolveValues(container, ...tokens) {
    return tokens.map((item) => resolveValue(container, item));
}
/**
 * Decorates a factory by passing resolved values as factory arguments.
 *
 * If an argument is an object which has tokens as its properties,
 * then returns an object containing resolved values as properties.
 *
 * @param factory - A factory.
 * @param tokens - Tokens which correspond to factory arguments.
 *
 * @return Decorated factory which takes a dependency container as a single argument.
 */
export function injectable(factory, ...tokens) {
    return (container) => {
        const values = resolveValues(container, ...tokens);
        return factory.apply(this, values);
    };
}
/**
 * Decorates a class by passing resolved values as arguments to its constructor.
 *
 * If an argument is an object which has tokens as its properties,
 * then returns an object containing resolved values as properties.
 *
 * @param constructor - Constructor of a class
 * @param tokens - Tokens which correspond to constructor arguments
 *
 * @return A factory function which takes a dependency container as a single argument
 * and returns a new created class.
 */
export function injectableClass(constructor, ...tokens) {
    return injectable((...values) => new constructor(...values), ...tokens);
}
//# sourceMappingURL=utils.js.map